/*
 * Copyright (C) 2018 SAP SE
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * Author: Michael Trapp <michael.trapp@sap.com>
 */

#ifndef __VIRTIO_H__
#define __VIRTIO_H__

/*
 * Initialize virtio layer
 */
int virtio_init(int max_channel, int expiration_period);

/*
 * Main virtio function
 * 'start_routine' of pthread_create()
 */
void *virtio_run(void *arg);

/*
 * Update the metrics response buffer of a VM/host
 */
int virtio_metrics_update(const char *buf,
                          int len,
                          int id,
                          const char *name);

/*
 * Stop virtio thread
 */
void virtio_stop(void);

#endif                          /* __VIRTIO_H__ */
